#!/usr/local/bin/perl
use strict;
use POSIX qw(strftime);
#
###########################################################################
# deductive and query conservativity:
#  see DL-08.tex for translation
# options to run with:
#  -d deductive conservativity 
#  -q query conservativity (must be either -d or -q)
#  -v verbose output (with all clauses in human-readable form)
#  -b batch-mode output (only numbers)
#  -s slicing (each atomic concept has a contiguous array of variables)
#  -sw the w_{j,k} are attached to the type 
#  -sq maximal type compactness (all variables for a type are in one place) 
#  -r prune TBox12 axioms (query conservativity only) 
#  -r1 simple TBox1 role axioms 
#  -r2 simple TBox12 role axioms 
###########################################################################

if ($ARGV[0] eq "")
{
    print STDERR <<ENDHELP;
TeX to QDIMACS translator (ver 0.6.1, 4 Feb 2008)
Usage:
   tex2qbf-5sf.pl options (file|directtory)
   
Options:
  -d deductive conservativity 
  -q query conservativity
  -v verbose output (with all clauses in LaTeX form)
  -b batch-mode output (m0, m1, m12, W0, W1, W12, A1, A12, C, V, VA, VE)
        m0, m1, m12 is the number of roles in TBox1, TBox12, Sigma
        W0, W1, W12 is the number of basic concepts in TBox1, TBox12, Sigma
        A1, A12 is the number of axioms in TBox1, TBox12
        C is the number of clauses
        V, VA, VE is the number of variables (total, universal, existential)
  -s slicing (each atomic concept has a contiguous array of variables)
  -sw the w_{j,k} are attached to the type 
  -sq maximal type compactness (all variables for a type are in one place) 
  -r TBox12 axiom pruning 
  -r1 simple TBox1 role axioms 
  -r2 simple TBox12 role axioms 
  
  If a file is specified then it is processed and QDIMACS is written to STDOUT.
  
  If a directory is specified then every .tex file of it is processed with the 
  specified options and their outputs are wtitten into the d-filename.qdimacs or
  q-filename.qdimacs files in the current directory (the extension depends on 
  the options).
ENDHELP

    exit;
}

my $now = strftime "%H:%M:%S on %d %b %Y", localtime;

my ($batch,$log,$slice,$deductive,$query,$pruneaxioms,$simpleroles1,$simpleroles12) = (0,0,0,0,0,0,0,0);

while ($ARGV[0] =~ /^-/)
{
    if ($ARGV[0] eq "-b") { $batch = 1;  } 
    elsif ($ARGV[0] eq "-v") { $log = 1; }
#    elsif ($ARGV[0] eq "-s") { $slice = 1; }
#    elsif ($ARGV[0] eq "-sw") { $slice = 2; }
#    elsif ($ARGV[0] eq "-sq") { $slice = 3; }
    elsif ($ARGV[0] eq "-d") { $deductive = 1; }
    elsif ($ARGV[0] eq "-q") { $query = 1; }
    elsif ($ARGV[0] eq "-r") { $pruneaxioms = 1; }
    elsif ($ARGV[0] eq "-r1") { $simpleroles1 = 1; }
    elsif ($ARGV[0] eq "-r2") { $simpleroles12 = 1; }
    else { die "ERROR: unknown option $ARGV[0]"; }
    shift @ARGV;
}

die "ERROR: either -d or -d must be used" if (!$deductive && !$query);
die "ERROR: both -d or -d cannot be used at the same time" if ($deductive && $query);
die "ERROR: -r can be used only with -q" if ($deductive && $pruneaxioms);

if (-d $ARGV[0] && -e $ARGV[0])
{
    my $dir = $ARGV[0];
    $dir =~ s/\/\z//;

    opendir(DIR, $dir) || die "can't opendir $dir: $!";
    my @filenames =  grep { !/^\./ && /\.tex\z/ && -f "$dir/$_" } readdir(DIR); #  
    closedir DIR;
    
    my $options = ($deductive ? "-d " : "-q ")
                 .(($slice == 1) ? "-s " : (($slice == 2) ? "-sw " : (($slice == 3) ? "-sq " : "")))
                 .($pruneaxioms ? "-r " : "")
                 .($simpleroles1 ? "-r1 " : "")
                 .($simpleroles12 ? "-r2 " : "");
    
    print STDERR "Options: $options\n";
    
    foreach (sort @filenames)
    {
        my $outfilename = $_;
        $outfilename =~ s/.tex\z/.qdimacs/;
        $outfilename =  ($deductive ? "d2-" : "q2-").$outfilename;
        print STDERR "$outfilename, ";
        system("$0 -b $options $dir/$_ >$dir/$outfilename") == 0 or die "system(\"$0 -b $options $dir/$_ >$dir/$outfilename\") failed: $?";
    }
    exit;
}
else
{
    die "ERROR: $ARGV[0] is not a file" if (!-f $ARGV[0] || !-e $ARGV[0]);
}

my $filenm = $ARGV[0];
$filenm =~ s/\\/\//g;

my @adir = split(/\//, $filenm);
pop(@adir);
my $dir = join("/", @adir);
$dir .= '/' if ($dir ne "");

open(FILE,$filenm) || die "can't open file $filenm: $!";
my @tex = <FILE>;        # Read the input into an array
close(FILE);

my $logConjunctiveRHS = 0;

my $curreq;
my (%concept0, %rn0);
my (@tbox1, @tbox12);
my (%concept1, %concept12);
my (%role1, %role12);
my ($len1, $len12, $ax1, $ax12, $conjclause1, $conjclause12) = (0,0,0,0,0,0);

my @V;
my (%av, %ev);

my %C;
my @postclauses;
my $gatherclauses = 0;

my %REPL;

#my %role1index;

my $inequationmode = 0;
my ($hasSigma, $hasTBox1, $hasTBox2) = (0,0,0);

my ($useU, $useV, $useE) = (0,0,0);

if ($deductive)
{  
    ($useU, $useV, $useE) = (1,1,1);
}
else
{
    ($useU, $useV, $useE) = (1,1,1);
}

$useE = 0 if ($simpleroles12);

foreach (@tex)
{
    s/%.*//; # remove comments
    
    if (/\\input\{([^\}]+)\}/)
    {
        my $ifn = $1;
        open(INCFILE,"$dir$ifn.tex") || die "can't open file $dir$ifn.tex: $!";
        my @inctex = <INCFILE>;        # Read the input into an array
        close(INCFILE);

        my $ss;
        foreach $ss (@inctex)
        {
            $ss =~ s/%.*//; # remove comments
        }
                
        my $incline = join(" ",@inctex);
        s/\\input\{$ifn\}/$incline/;
    }

    if (/\\begin{align\*?}/)
    {
        $_ = $';
        $inequationmode = 1;
    }
    
    if ($inequationmode)
    {
        if (/\\end{align\*?}/)
        {
            $curreq = normaliseEquation($curreq.$`);
            $curreq =~ /([^=]+)=\s*\\{(.*)\\}/;  # split into name and axioms

            if(!$hasSigma)
            {
                print STDERR "Processing \Sigma\n" if (!$batch);
                my $name = $1;

                my ($c,$r) = split(/;/,$2);

                print "c ".($deductive ? "Deductive" : "Query")." conservative extension QBF (slicing mode: $slice, TBox1 simple roles: $simpleroles1, TBox12 simple roles: $simpleroles12, TBox12 axiom pruning: $pruneaxioms)\n";
                print "c    (generated at $now by tex2qbf-5sf.pl of 4 Feb 2008)\nc\n";
                
                print "c Signature concepts: ";
                foreach (split(/,/,$c))
                {
                    s/ //g; # remove spaces
                    next unless ($_);
                    $concept0{$_} = 1;
                    print "$_ ";
                    $concept1{$_} = 1; # the concept is in \Sigma, add to T_1 (it will be added to T_{12} later)
                }
                print "\nc Signature roles: ";
                foreach (split(/,/,$r))
                {
                    s/ //g; # remove spaces
                    next unless ($_);
                    $rn0{$_} = 1;
                    print "$_ ";
                    $role1{$_}{1} = 1; # the role is in \Sigma, add to  T_1 (it will be added to T_{12} later)
                    $role1{"$_^{-}"}{1} = 1; #
                }
                print "\nc \n";
                $hasSigma = 1;
            }
            elsif (!$hasTBox1)
            {
                my $name = $1;
                @tbox1 = split(/,/,$2);

                ($len1, $ax1, $conjclause1) = parseTBox($name,\@tbox1,\%concept1,\%role1);
                $hasTBox1 = 1;
            }
            elsif (!$hasTBox2)
            {
                my $name = $1;
                @tbox12 = @tbox1;
                push(@tbox12, split(/,/,$2));

                foreach (keys %role1)
                {
                    %{ $role12{$_} } = %{ $role1{$_} };   # copy parameters to Q_{12}
                }

                foreach (keys %concept1)
                {
                    $concept12{$_} = $concept1{$_};   # copy concepts
                }

                ($len12, $ax12, $conjclause12) = parseTBox($name,\@tbox12,\%concept12,\%role12);

                print STDERR "\n" if (!$batch);

                ########################################################

                foreach (keys %role1)
                {
                    %{ $role1{$_} } = %{ $role12{$_} };   # make parameters Q_{12}
                    my $cl = generateNumberRestrictionsAxioms(\@tbox1, $_, keys %{ $role1{$_} });
                    $len1 += 2 * $cl;
                    $ax1 += $cl;
                }

                print "c Extended TBox1 axioms: ($ax1 axioms of overall length $len1 (plus $conjclause1))\n";
                printTBox(\@tbox1);

                ########################################################

                foreach (keys %role12)
                {
                    my $cl = generateNumberRestrictionsAxioms(\@tbox12, $_, keys %{ $role12{$_} });
                    $len12 += 2 * $cl;
                    $ax12 += $cl;
                }

                print "c Extended TBox12 axioms: ($ax12 axioms of overall length $len12  (plus $conjclause12))\n";
                printTBox(\@tbox12);
                
                ########################################################

                if ($pruneaxioms)
                {
                    foreach (@tbox12)
                    {
                        my $intbox1 = 0;
                        my $tb1;
                        foreach $tb1 (@tbox1)
                        {
                            $intbox1 = 1 if ($tb1 eq $_);
                        }
                        next if (!$intbox1);
                    
                        my ($inSigma,$ax122,$len122,$conjclause122) = (1,0,0,0);
                        my ($lhs,$rhs) = split(/\\sqsubseteq/);

                        next if (!$rhs);

                        my @L = split(/\\sqcap/,$lhs);

                        if ($rhs =~ /\\sqcap/) # conjunctive right-hand side
                        {
                            $conjclause122 = -1;
                            foreach (split(/\\sqcap/,$rhs))
                            {
                                my $c = getConcept($_); 
                                die "\\bot not allowed in the rhs of $lhs -> $rhs" if ($c eq "");
                                $inSigma = 0 if (!inSignature($_,\%concept0,\%rn0));
                                ++$conjclause122;
                            }
                            ++$len122;
                        }
                        else # if not conjunctive
                        {
                            push(@L, split(/\\sqcup/,$rhs));
                        }

                        foreach (@L)
                        {
                            my $c = getConcept($_); 
                            next if ($c eq "");
                            # handling of \bot
                            $inSigma = 0 if (!inSignature($_,\%concept0,\%rn0));
                            ++$len122;
                        }
                        
                        if ($inSigma)
                        {
                            print STDERR "ERROR PRUNE: $_\n" if (!$intbox1);
                            $_ = "(pruned $lhs -> $rhs)";
                            --$ax12;
                            $len12 -= $len122;
                            $conjclause12 -= $conjclause122;
                            print STDERR "PRUNE $lhs \\sqsubseteq $rhs of length $len122 ($conjclause122)\n" if (!$batch); 
                        }
                    }
                    # TODO: handling of empty TBox2
                    print "c Pruned TBox12 axioms: ($ax12 axioms of overall length $len12  (plus $conjclause12))\n";
                    printTBox(\@tbox12);
                }

                ########################################################

                createFullListOfBasicConcepts (\%concept1, \%role1);
                createFullListOfBasicConcepts (\%concept12, \%role12);

                ########################################################

                my $m0 = keys %rn0;
                
                my ($m1, $m12) = ((keys %role1) / 2, (keys %role12) / 2);

                my $B1 = 0; # the index of the first type for TBox1 (indices start with 0)
                my $T1 = $m1; # the index of the last type for TBox1 
                my $N1 = $T1 - $B1 + 1; # or the number of types for TBox1
                
                my $B12 = $T1 + 1; # the index of the first type for TBox12 
                my $TE12 = $B12 + (($deductive) ? 0 : $m1);  ## end of extensions 

                my $T12 = ($deductive) ? ($B12 + $m12) : ($B12 + $m1 + $m12 - $m0); # the index of the last type for TBox12 
                my $N12 = $T12 - $B12 + 1; # the number of `syntactic' types in TBox12 
                
                ## start of copy-types (for query conservativity, it is outside the final type)
                my $BC12 = (!$deductive) ? ($TE12 + 1) : ($T12 + 1);  
                my $NC12 = (!$deductive) ? ($T12 - $BC12 + 1) : 0; 
                ## start of full-types (for deductive conservativity; for query, it is outside the final type)
                my $BF12 = ($deductive) ? ($TE12 + 1) : ($T12 + 1);  

                my $rolenumber12 = ($deductive) ? $m12 : ($m12 - $m0);


                ########################################################################
                # types layout
                # deductive: T1 types: B1-T1 (N1 is the number), 
                #            T12 types: extensions B12-TE12, full-types BF12-T12
                # query:     T1 types: B1-T1 (N1 is the number)
                #            T12 types: extensions B12-TE12, copy-types BC12-T12 
                ######################################################################

                print STDERR "Role names: $m1 in TBox1, $m12 in TBox12, $m0 in Sigma\n" if (!$batch);

                ########################################################

                my ($w1, $w12, $w0) = (0,0,0); # number of basic concepts in T_1, T_12, \Sigma

                foreach (keys %concept1)
                {
                    ++$w1;
                }

                foreach (keys %concept12)
                {
                    if (inSignature($_,\%concept0,\%rn0))
                    {
                        ++$w0;
                        $concept0{$_} = 1;   
                    }
                    ++$w12;
                }

                print STDERR "Basic concepts: $w1 in TBox1, $w12 in TBox12, $w0 in Sigma\n" if (!$batch);

                #######################################################
                # TBox 1
                #######################################################

                my $base = 0;
                my $width1 = $w1;
                $width1 += $ax1 if ($slice == 2);

                {
                    my $c = 0;
                    foreach (sort keys %concept1)
                    {
                        for (my $j = $B1; $j <= $T1; ++$j)
                        {
                            my $v = $base + (($slice == 1) ? ($N1 * $c + $j + 1) : ($width1 * $j + $c + 1));
                            $C{"($_)^{j}"}{$j} = $v;
                            $av{$v} = 1;
                        }
                        ++$c;
                    }
                    $base += $N1 * $width1;
                }

                #########################################################
                # s, u, v
                #########################################################

                my @psi;

                {
                    for (my $j = 1; $j <= $m1; ++$j)
                    {
                        my $vs = $base + $j;
                        $C{'\s_{j}'}{$j} = $vs;
                        $ev{$vs} = 1;
                        push(@psi, "-$vs");

                        if ($useU)
                        {
                            my $vu = $base + $m1 + $j;
                            $C{'\u_{j}'}{$j} = $vu;
                            $ev{$vu} = 1;
                            push(@psi, "-$vu");
                        }

                        if ($useV)
                        {
                            my $vv = $base + 2*$m1 + $j;
                            $C{'\v_{j}'}{$j} = $vv;
                            $ev{$vv} = 1;
                            push(@psi, "-$vv");
                        }
                    }
                    $base += $m1;
                    $base += $m1 if ($useU);
                    $base += $m1 if ($useV);
                }


                #########################################################
                # w
                #########################################################
                
                {
                    for (my $k = 1; $k <= $ax1; ++$k)
                    {
                        for (my $j = $B1; $j <= $T1; ++$j)
                        {
                            my $v = ($slice == 1) ? ($base + $N1 * ($k - 1) + $j + 1) 
                                 : (($slice == 0) ? ($base + $ax1 * $j + $k) 
                                                  : ($w1 + $width1 * $j + $k));
                            $C{"\\w_{$k,j}"}{$j} = $v;
                            $ev{$v} = 1;
                            push(@psi, "-$v");
                        }
                    }
                    $base += $N1 * $ax1 if (($slice == 0) || ($slice == 1));
                }

                ########################################################
                # p
                ########################################################
                {
                    my $v = $base + 1;
                    $C{"\\p"}{0} = $v;
                    $ev{$v} = 1;
                    ++$base;
                }


                ########################################################
                # g 
                ########################################################
                
                if (!$deductive)
                {
                    for (my $k = $BC12; $k <= $T12; ++$k)
                    {
                        my $qk = "\\q_{$k,j}";
                        for (my $j = $B1; $j <= $T1; ++$j)
                        {
                            my $v = $base + ($k - $BC12) * $N1 + $j + 1;
                            $C{$qk}{$j} = $v;
                            $ev{$v} = 1;
                        }
                    }
                    $base += $N1 * $NC12;
                }


                #########################################################
                # TBox 12 types
                #########################################################

                my %tp12;
                for (my $j = $B12; $j <= $TE12; ++$j)  # extensions
                {
                    $tp12{$j} = "-$C{'\p'}{0}";
                }
                for (my $j = $BF12; $j <= $T12; ++$j) # full-types
                {
                    $tp12{$j} = "-$C{'\p'}{0}";
                }
                for (my $k = $BC12; $k <= $T12; ++$k) # copy-types
                {
                    my $qk = "\\q_{$k,j}";
                    for (my $j = $B1; $j <= $T1; ++$j)
                    {
                        $tp12{"$k,$j"} = "-$C{$qk}{$j}";
                    }
                }

                #########################################################
                # TBox 12
                #########################################################

                {
                    my $w = ($slice != 3) ? (($deductive) ? $w12 : $w12) : $width1;
                    
                    my $c = 0;
                    foreach (sort keys %concept12)
                    {
                        my $inSigma = inSignature($_,\%concept0,\%rn0);
                        
                        foreach my $jj (sort keys %tp12) #for (my $j = $B12; $j <= $T12; ++$j)
                        {
                            my ($j12, $j0) = ($jj =~ /(\d+),(\d+)/) ? ($1, $2) : ($jj, $jj);
                            
                            if (!$inSigma || ($j12 >= $BF12))
                            {
                                my $v = ($slice == 1) ? ($base + $N12 * $c + ($j12 - $B12) + 1) : 
                                           (($slice != 3) ? ($base + $w * ($j12 - $B12) + $c + 1) 
                                                          : ($w1 + $ax1 + $w * ($j12 - $B12) + $c + 1));
                                $C{"($_)^{j}"}{$jj} = $v;

                                $ev{$v} = 1;
                            }
                            else
                            {
                                # if not a copy-type (for query case) then is in sigma and it is an extension type 
                                $C{"($_)^{j}"}{$jj} = ($j12 >= $BC12) ? $C{"($_)^{j}"}{$j0} : $C{"($_)^{j}"}{$j12 - $B12};
                            }
                        }
                        ++$c;
                    }
                    $base += $N12 * $w if ($slice != 3);
                }


                ########################################################
                # e
                ########################################################
                
                if ($useE)
                {
                    for (my $j = 1; $j <= $rolenumber12; ++$j)
                    {
                        my $v = $base + $j;
                        $C{"\\e_{j}"}{$j} = $v;
                        $ev{$v} = 1;
                    }
                    $base += $m12;
                }

                #########################################################
                # print out symbol table and fill in the inverse search
                #########################################################
                {
                    my $c;
                    foreach $c (sort keys %C)
                    {
                        my @vars = values %{$C{$c}};
                        print "c $c:";
                        foreach (sort {$a <=> $b} keys %{$C{$c}})
                        {
                            print " $C{$c}{$_}";
                            my $cj = $c;
                            $cj =~ s/j}/$_}/;
                            if ($V[$C{$c}{$_}])
                            {
                                my $cjb = $c;
                                my $jb = $_ - $B12;
                                $cjb =~ s/j}/$jb}/;
                                my $ccjb = $cj;
                                $ccjb =~ s/,\d+}/}/;
                                my $cccjb = $V[$C{$c}{$_}];
                                $cccjb =~ s/,\d+}/}/;
                                my $ccccjb = $cj;
                                $ccccjb =~ s/\^{\d+,/\^{/;
                                print STDERR "INTERNAL ERROR: $V[$C{$c}{$_}] and $cj are the same variable!!! ($C{$c}{$_},$cjb, $ccjb)\n" 
                                            if (($V[$C{$c}{$_}] ne $cjb) && ($cccjb ne $ccjb) && ($V[$C{$c}{$_}] ne $ccccjb));
                            }
                            else
                            {
                                $cj =~ s/,\d+}/}/ unless (($cj =~ /\\q/) || ($cj =~ /\\w/));
                                $V[$C{$c}{$_}] = $cj;
                            }
                        }
                        print "\n";
                    }
                }
                print "c ---\nc\n";
                if ($log)
                {
                    for(my $k = 1; $k < @V; ++$k)
                    {
                        print "c $k: $V[$k]\n";
                    }
                    print "c ---\nc\n";
                }

                #########################################################

                my ($out1, $out12) = (0,0); 
                my ($ctg1, $ctg12) = (($m1 + 1)*$len1, (($deductive) ? ($m12 + 1) : (keys %tp12))*($ax12 + $conjclause12));
                
                $ctg1 += 1 + (($simpleroles1) ? (2*$m1*($m1 + 2) + 3*$m1) : (2*$m1*($m1 + 1) + 2*$m1)); 
                $ctg12 += $NC12 if (!$deductive);
                
                $ctg12 += $rolenumber12*(2 + 2*(($simpleroles12) ? ($N12 - 1) : $N12));

                #########################################################

                print "p cnf ".(@V - 1)." ".($ctg1 + $ctg12)."\n";
                my @ava = sort {$a <=> $b} (keys %av);
                print "a @ava 0\n";
                my @eva = sort {$a <=> $b} (keys %ev);
                print "e @eva 0\n";

                #########################################################
                # psi clause
                #########################################################

                $out1 += outClause("$C{'\p'}{0}", @psi);

                ###########################################################
                # TBox 1 roles
                ###########################################################

                {
                    my $i = $B1 + 1;
                    foreach (keys %role1)
                    {
                        next if (/\^{-}/);  ## skip inverse roles

                        my ($r, $ri) = ("(\\exists $_)^{j}", "(\\exists $_^{-})^{j}");

                        my (@rline);
                        for (my $j = $B1; $j <= $T1; ++$j)
                        {
                            $out1 += outClause("-$C{$r}{$j}", "$C{'\u_{j}'}{$i}");
                            $out1 += outClause("-$C{$ri}{$j}", "$C{'\v_{j}'}{$i}");
                            push(@rline, "$C{$r}{$j}");
                            push(@rline, "$C{$ri}{$j}");
                        }

                        if ($simpleroles1)
                        {
                            $out1 += outClause("$C{$ri}{$i}", "$C{'\u_{j}'}{$i}");
                            $out1 += outClause("$C{$r}{$i}", "$C{'\v_{j}'}{$i}");
                            $out1 += outClause("-$C{$r}{$i}", "$C{'\s_{j}'}{$i}");
                            $out1 += outClause("-$C{$ri}{$i}", "$C{'\s_{j}'}{$i}");
                        }
                        else
                        {
                            $out1 += outClause("-$C{'\u_{j}'}{$i}", "-$C{'\v_{j}'}{$i}", "$C{'\s_{j}'}{$i}");
                        }
                        $out1 += outClause(@rline, "$C{'\s_{j}'}{$i}");

                        ++$i;
                    }
                    die "INTERNAL ERROR: $i is not the same as $T1" if ($i - 1 != $T1);
                }

                ###########################################################
                # TBox 1 axioms
                ###########################################################

                #$gatherclauses = 1 if ($clauseorder >= 1);

                my $k = 0;
                foreach (@tbox1)
                {
                    my($lhs,$rhs) = split(/\\sqsubseteq/);

                    next if (!$rhs); # skip if the axiom is empty

                    die "ERROR: No \\sqcup is allowed in LHS of $lhs \\sqsubseteq $rhs\n" if ($lhs =~ /\\sqcup/);
                    
                    ++$k; # the number of the current axiom
                    my $wk = "\\w_{$k,j}";

                    my $emptyLout = 1;

                    my @L = split(/\\sqcap/,$lhs);

                    if ($rhs =~ /\\sqcap/) # conjunctive right-hand side
                    {
                        my @crhline;

                        print "conjunctive axiom: $lhs \\sqsubseteq $rhs\n" if $logConjunctiveRHS;
                        foreach (split(/\\sqcap/,$rhs))
                        {
                            my $nc = getConceptSign1();
                            my $c = getConcept ($_);
                            print "conjunctive rhs concept: $_ $c\n" if $logConjunctiveRHS;

                            die "ERROR: No \\top or \\bot allowed in conjunctive RHS of $lhs \\sqsubseteq $rhs\n" unless ($c);
                            
                            for (my $j = $B1; $j <= $T1; ++$j)
                            {
                                push(@{$crhline[$j]}, "$nc$C{$c}{$j}");
                            }
                            $emptyLout = 0;
                        }

                        for (my $j = $B1; $j <= $T1; ++$j)
                        {
                            $out1 += outClauseJ($j,@{$crhline[$j]}, "$C{$wk}{$j}");
                        }
                    }
                    else
                    {
                        push(@L, negate(split(/\\sqcup/,$rhs)));
                    }

                    #print "axiom: $lhs \\sqsubseteq $rhs\n";
                    #print "LHS: @L\n";

                    foreach (@L)
                    {
                        my $nc = getConceptSign12();
                        my $c = getConcept ($_);

                        for (my $j = $B1; $j <= $T1; ++$j)
                        {
                            if ($c)
                            {
                                $out1 += outClauseJ($j,"$nc$C{$c}{$j}", "$C{$wk}{$j}");
                                $emptyLout = 0;
                            }
                            elsif ($nc ne "-") # if \bot (skip if \top)
                            {
                                $out1 += outClauseJ($j,"$C{$wk}{$j}");
                                $emptyLout = 0;
                            }
                        }
                    }
                    if ($emptyLout) # LHS is equivalent to \top, RHS is \bot = TBox is unsatisfiable
                    {
                        print STDERR "\n\n\nWARNING: TBox1 is UNSATISFIABLE\n\n\n";
                        for (my $j = $B1; $j <= $T1; ++$j)
                        {
                            $out1 += outClauseJ($j,"-$C{$wk}{$j}");
                        }
                    }
                }
                
                #$out1 += flushClauses($t) if ($clauseorder >= 1);

                #print STDERR "TBox1 out: $out1 (of $ctg1)\n" if (!$batch);

                ###########################################################
                # TBox 12 roles
                ###########################################################

                {
                    my $i = 0;
                    my @roletypes = grep { !((/(\d+),(\d+)/) && ($2 ne 0)) } (sort keys %tp12);
                    #print "ROLE TYPES: @roletypes\n";
                    foreach (keys %role12)
                    {
                        next if (/\^{-}/);  ## skip inverse roles

                        #print "ROLE $_\n";
                        if (!$deductive && $rn0{$_}) ## in signature
                        {
                            print STDERR "$_ is in signature\n" if (!$batch);
                            next;
                        }                    

                        my ($r, $ri) = ("(\\exists $_)^{j}", "(\\exists $_^{-})^{j}");
                        
                        my (@rline, @riline);
                        my $j;
                        foreach $j (@roletypes) #(my $j = $BC12; $j <= $TC12; ++$j)
                        {
                            push(@rline, "$C{$ri}{$j}");
                            push(@riline, "$C{$r}{$j}");
                        }
                        
                        if(!$simpleroles12)
                        {
                            my $ii = $i + 1;
                            $out12 += outClause(@rline, "$C{'\e_{j}'}{$ii}", "-$C{'\p'}{0}");
                            $out12 += outClause(@riline, "$C{'\e_{j}'}{$ii}", "-$C{'\p'}{0}");

                            foreach my $j (@roletypes) 
                            {
                                $out12 += outClause("-$C{'\e_{j}'}{$ii}", "-$C{$r}{$j}");
                                $out12 += outClause("-$C{'\e_{j}'}{$ii}", "-$C{$ri}{$j}");
                            }
                        }
                        else
                        {
                            $out12 += outClause(@rline, "-$C{$r}{$roletypes[$i]}", "-$C{'\p'}{0}");
                            $out12 += outClause(@riline, "-$C{$ri}{$roletypes[$i]}", "-$C{'\p'}{0}");

                            foreach my $j (@roletypes) 
                            {
                                next if ($j eq $roletypes[$i]);
                                $out12 += outClause("$C{$r}{$roletypes[$i]}", "$C{$ri}{$roletypes[$i]}", "-$C{$r}{$j}", "-$C{'\p'}{0}");
                                $out12 += outClause("$C{$r}{$roletypes[$i]}", "$C{$ri}{$roletypes[$i]}", "-$C{$ri}{$j}", "-$C{'\p'}{0}");
                            }
                        }                        
                        
                        ++$i;
                    }
                    #die "INTERNAL ERROR: $i is not $T12" if ($simpleroles12 && ($i != $N12));
                }

                {
                    my $cctg12 = 0; # (($deductive) ? ($m12 + 1) : (keys %tp12))*($ax12 + $conjclause12);
                    $cctg12 += $rolenumber12*(2 + 2*(($simpleroles12) ? ($N12 - 1) : $N12));
                
                    #$cctg12 += $NC12 if (!$deductive);
                    die "INTERNAL ERROR: $out12 and $cctg12 after TBox12 roles $N12 types and $rolenumber12 roles" if ($out12 != $cctg12);
                }

                ###########################################################
                # TBox 12 chi
                ###########################################################

                if (!$deductive)
                {
                    for (my $k = $BC12; $k <= $T12; ++$k)
                    {
                        my $qk = "\\q_{$k,j}";
                        my @qline;
                        for (my $j = $B1; $j <= $T1; ++$j)
                        {
                            push(@qline,"$C{$qk}{$j}");
                            foreach (sort keys %concept0)
                            {
                                my $c = getConcept($_);
                                setReplacement("$C{$c}{$k}", "$C{$c}{$j}", "-$C{$qk}{$j}");
                            }
                            #foreach (sort keys %concept0)
                            #{
                            #    my $c = getConcept($_);
                            #    $out12 += outClause("-$C{$qk}{$j}", "-$C{$c}{$j}", "$C{$c}{$k}");
                            #    $out12 += outClause("-$C{$qk}{$j}", "$C{$c}{$j}", "-$C{$c}{$k}");
                            #}
                            ##for (my $jj = $j + 1; $jj <= $T1; ++$jj)
                            ##{
                            ##    $out12 += outClause("-$C{$qk}{$j}", "-$C{$qk}{$jj}");
                            ##}
                        }
                        $out12 += outClause(@qline, "-$C{'\p'}{0}");
                    }
                }

                {
                    my $cctg12 = 0; # (($deductive) ? ($m12 + 1) : (keys %tp12))*($ax12 + $conjclause12);
                    $cctg12 += $rolenumber12*(2 + 2*(($simpleroles12) ? ($N12 - 1) : $N12));
                
                    $cctg12 += $NC12 if (!$deductive);
                    die "INTERNAL ERROR: $out12 and $cctg12" if ($out12 != $cctg12);
                }


                ###########################################################
                # TBox 12 axioms
                ###########################################################

                #$gatherclauses = 1 if ($clauseorder == 2);

                my @type12list = sort keys %tp12;

                my $k12 = 0;
                foreach (@tbox12)
                {
                    my($lhs,$rhs) = split(/\\sqsubseteq/);
                    my(%lcl,%rcl);

                    next if (!$rhs);

                    ++$k12;

                    my @L = split(/\\sqcap/,$lhs);

                    if ($rhs =~ /\\sqcap/) # conjunctive right-hand side
                    {
                        print "conjunctive axiom 2: $lhs \\sqsubseteq $rhs\n" if $logConjunctiveRHS;
                        foreach (split(/\\sqcap/,$rhs))
                        {
                            my $nc = getConceptSign12();
                            my $c = getConcept($_);
                            print "conjunctive rhs concept: $_ $c\n" if $logConjunctiveRHS;

                            die "ERROR: No \\top or \\bot allowed in conjunctive RHS of $lhs \\sqsubseteq $rhs\n" unless ($c);
                            
                            foreach (@type12list) 
                            {
                                push(@{$rcl{$_}}, "$nc$C{$c}{$_}");
                            }
                        }
                    }
                    else # if not conjunctive
                    {
                        push(@L, negate(split(/\\sqcup/,$rhs)));

                        foreach (@type12list) 
                        {
                            push(@{$rcl{$_}}, "\\bot");
                        }
                    }

                    #print "axiom: $lhs \\sqsubseteq $rhs\n";
                    #print "LHS 12: @L\n";

                    my $botLHS = 0;

                    foreach (@L)
                    {
                        my $nc = getConceptSign1();
                        my $c = getConcept ($_);
                        #print "$_ $c\n";

                        if ($c)
                        {
                            foreach (@type12list) 
                            {
                                push(@{$lcl{$_}}, "$nc$C{$c}{$_}");
                            }
                        }
                        elsif ($nc eq "-") # if \bot (ignore if \top on LHS)
                        {
                            $botLHS = 1;
                        }
                    }

                    if (!$botLHS)
                    {
                        foreach (@type12list) 
                        {
                            foreach my $r (@{$rcl{$_}})
                            {
                                $out12 += outClauseJ($_, @{$lcl{$_}}, $r, $tp12{$_});
                            }
                        }
                    }
                    else
                    {
                        print STDERR "WARNING: Axiom $lhs \\sqsubseteq $rhs is REDUNDANT\n";
                    }
                }

                #$out12 += flushClauses($N) if ($clauseorder == 2);

                #print STDERR "TBox12 out: $out12 (of $ctg12)\n" if (!$batch);

                die "ERROR: wrong number of clauses for TBox 1 has been generated ($out1 instead of $ctg1)" if ($out1 != $ctg1);
                die "ERROR: wrong number of clauses for TBox 12 has been generated ($out12 instead of $ctg12)" if ($out12 != $ctg12);
                print STDERR "total out: ".($out1 + $out12)." clauses with ".(@V-1)." variables (A: ".@ava.", E: ".@eva.")\n" if (!$batch); # @V starts with index 1

                print STDERR "$m0, $m1, $m12, $w0, $w1, $w12, $ax1, $ax12, ".($out1 + $out12).", ".(@V-1).", ".@ava.", ".@eva."\n" if ($batch);
                
                $hasTBox2 = 1;
            }
            else 
            {
                print STDERR "WARNING: too many equations\n";
            }

            $inequationmode = 0;
            $curreq = "";
        }
        else
        {
            $curreq .= $_;
        }
    }
}


sub setReplacement
{
    my $w = shift;
    push(@{$REPL{$w}}, [@_]);
    #print STDERR "REPLACEMENT: $w to (@_)\n";
}

sub normaliseEquation ($)
{
    my ($eq) = @_;

    $eq =~ s/_([^{])/_{\1}/g; # put every subscript in brackets
    $eq =~ s/\^([^{])/\^{\1}/g; # put every superscript in brackets
    $eq =~ s/\^({[^}]*})_({[^}]*})/_\2\^\1/g; # put every subscript before superscript
    
    $eq =~ s/&/ /g; # remove column separators
    $eq =~ s/\\\\/ /g; # remove \\
    $eq =~ s/~/ /g; # remove ~
    $eq =~ s/\n/ /g; # remove new lines
    $eq =~ s/\\exists/\\mathop{\\geq 1}/g; # replace \exists with \geq 1
    $eq =~ s/\\ge[^a-zA-Z]/\\geq/g; # expand \ge to \geq
    $eq =~ s/\\le[^a-zA-Z]/\\leq/g; # expand \le to \leq
    $eq =~ s/\\geq/\\geq /g; # add extra space after \geq
    $eq =~ s/\\leq/\\leq /g; # add extra space after \leq
    $eq =~ s/\\top/\\neg \\bot/g; # replace \top

    while ($eq =~ s/  / /g) # compress spaces
    {
    }

    $eq =~ s/\\top ?\\sqcap//g; # remove \top \sqcap
    $eq =~ s/\\sqcap ?\\top//g; # remove \sqcap \top
    $eq =~ s/\\neg ?\\top ?\\sqcup//g; # remove \bot \sqcup
    $eq =~ s/\\sqcup ?\\neg ?\\top//g; # remove \sqcup \bot

    $eq =~ s/\\mathop{\\leq\s?(\d+)}/"\\neg \\mathop{\\geq ".($1+1)."}"/ge; # replace \leq with \geq

    while ($eq =~ s/\\neg ?\\neg//g) # remove double negations
    {
    }
    while ($eq =~ s/  / /g) # compress spaces one again
    {
    }

    # print STDERR "equation: $curreq\n";
    $eq;
}


sub getConceptSign1
{
    if (/\\neg/) # strip off \neg
    {
        $_ = $';
        ""; # return
    }
    else
    {
        "-"; # return
    }
}


sub getConceptSign12
{
    if (/\\neg/) # strip off \neg
    {
        $_ = $';
        "-"; # return
    }
    else
    {
        ""; # return
    }
}

sub flushClauses
{
    my $t = shift;
    my $out = 0;
    for(my $j = 0; $j <= $t; ++$j)
    {
        foreach (@{$postclauses[$j]})
        {
            $out += outClause(@{$_});
        }
    }
    $gatherclauses = 0;
    @postclauses = undef;
    $out;
}


sub outClauseJ
{
    my $j = shift;
    if ($gatherclauses)
    {
        push(@{$postclauses[$j]}, [@_]);
        0;
    }
    else
    {
        outClause(@_);
    }
}

sub outClause1
{
    my @tbr = grep { /\A-?(\d+)\z/ && $REPL{$1} } @_;
    foreach (@tbr)
    {
        s/-//;
    }
    
    if (@tbr)
    {
        #print "TO BE REPLACED: @tbr in ";
        #outClause0(@_);        

        my $rr;
        foreach $rr (@{$REPL{$tbr[0]}}) # enumerate types
        {
            my ($rc, $rq) = @{$rr}; # $rq is the current type
            my %rmap;
            $rmap{$C{'\p'}{0}} = $rq;
            $rmap{$C{'\p'}{0}} =~ s/-//;
            
            foreach (@tbr)
            {
                my $tt;
                foreach $tt (@{$REPL{$_}}) 
                {
                    $rmap{$_} = ${@{$tt}}[0] if(${@{$tt}}[1] eq $rq);
                }
            }

            #print " REPLACEMENT MAP: {";
            #foreach (keys %rmap)
            #{
            #    print "$_ => $rmap{$_}, ";
            #}
            
            my @cp = @_;
            foreach (@cp)
            {
                my $cv = $_;
                $cv =~ s/-//;
                my $cc;
                foreach $cc (keys %rmap)
                {
                    s/$cc/$rmap{$cc}/ if ($cv == $cc);
                }
            }
        
            #print "} RESULT: ";
            outClause0(@cp);
        }
        1;
    }
    else
    {
        outClause0(@_);
    }
}

sub outClause
{
    outClause0(@_);
}

my $clausenumber = 0;

sub outClause0
{
    ++$clausenumber;
    my $s = join(" ", @_);
    $s =~ s/\\bot//g;
    print "[$clausenumber] " if $log;
    print "$s 0";
    print "\n" unless $log;

    my @t;
    foreach (@_)
    {
        next if ($_ eq "\\bot");
        
        die "INTERNAL ERROR: null var in @_" if ($_ == 0);
        if (/-/)
        {
            push(@t,"\\neg $V[$']");
        }
        else
        {
            push(@t,"$V[$_]");
        }
    }
    $s = join(" \\lor ", @t);
    print "  ($s)\n" if $log;

    1;
}


sub negate
{
    my @t;
    foreach (@_)
    {
        push(@t, ((/\\neg/) ? $' : "\\neg$_"));
    }
    @t;
}

sub createFullListOfBasicConcepts (\%\%)
{
    my ($concept,$role) = @_;

    my $r;
    foreach $r (keys %$role)
    {
        foreach (keys %{$$role{$r}})
        {
            if ($_ == 1)
            {
                $$concept{"\\exists $r"} = ${%{$$role{$r}}}{$_};
            }
            else
            {
                $$concept{"\\mathop{\\geq $_} $r"} = ${%{$$role{$r}}}{$_};
            }
        }
    }
}

sub getConcept
{
    my $c = shift;
    $c =~ s/\\exists/\\mathop{\\geq 1}/;

    if ($c =~ /\\mathop{\\geq\s?(\d+)}/)
    {
        my ($q,$r) = ($1, $');
        $r =~ s/ //g; # remove spaces

        ($q == 1) ? "(\\exists $r)^{j}" : "(\\mathop{\\geq $q} $r)^{j}";
    }
    else
    {
        $c =~ s/ //g; # remove spaces
        ($c eq '\bot') ? "" : "($c)^{j}";
    }
}



sub inSignature ($\%\%)
{
    my ($c,$concept,$rn) = @_;
    my $i = 0;

    $c =~ s/\\exists/\\mathop{\\geq 1}/g;

    if ($c =~ /\\mathop{\\geq\s?(\d+)}/)
    {
        $c = $';
        $c =~ s/ //g; # remove spaces
        $c =~ s/\^{-}//;
        $i = $$rn{$c};
    }
    else
    {
        $c =~ s/ //g; # remove spaces
        $i = $$concept{$c} unless ($c eq '\bot');
    }
    $i;
}


sub printTBox (\@)
{
    my ($tbox) = (@_);
    foreach (@$tbox)
    {
        print "c $_\n";
    }
    print "c\n";
}

sub parseTBox ($\@\%\%)
{
    my ($name,$tbox,$concept,$role) = @_;
    my ($len, $ax, $conjclause) = (0,0,0);

    print STDERR "TBox name: $name\n" if (!$batch);
    print STDERR "TBox axioms:\n" if (!$batch);
    foreach (@$tbox)
    {
        print STDERR "  $_\n" if (!$batch);
    }

    foreach (@$tbox)
    {
        my($lhs,$rhs) = split(/\\sqsubseteq/);

        if (!$rhs)
        {
            if ($_ ne " ")
            {
                die "ERROR: no \\sqsubseteq in ($_)";
            }
        }
        else
        {
            foreach (split(/\\sqcap/,$lhs))
            {
                $len += parseConcept($concept,$role,$_);
            }

            if ($rhs =~ /\\sqcap/)
            {
                print "conjunctive axiom: $lhs \\sqsubseteq $rhs\n" if $logConjunctiveRHS;
                foreach (split(/\\sqcap/,$rhs))
                {
                    $conjclause += parseConcept($concept,$role,$_);
                }
                ++$len;
                --$conjclause;
            }
            else
            {
                foreach (split(/\\sqcup/,$rhs))
                {
                    $len += parseConcept($concept,$role,$_);
                }
            }

            ++$ax;
        }
    }

    ($len, $ax, $conjclause); # return value
}

sub parseConcept (\%\%$)
{
    my($concept,$role,$c) = @_;

    $_ = $c;
    $_ = $' if (/\\neg/);

    if (/\\mathop{\\geq\s?(\d+)}/)
    {
        my ($q,$r) = ($1, $');
        $r =~ s/ //g; # remove spaces

        my ($ir) = (($r =~ /([^^]*)\^{-}/) ? $1 : "$r^{-}");

        #print "role: $r $ir $q\n";

        ${%{$$role{$r}}}{$q} = 1;
        ${%{$$role{$r}}}{1} = 1;
        ${%{$$role{$ir}}}{1} = 1;
    }
    else
    {
        s/ //g; # remove spaces
        $$concept{$_} = 1 unless ($_ eq '\bot');
    }

    ($_ ne '\bot'); # return value
}

sub generateNumberRestrictionsAxioms (\@$@)
{
    my ($tbox, $rn, @q) = @_;

    #print "$rn: @q\n";

    my $qs = 0;
    foreach (sort {$a <=> $b} @q)
    {
        if ($qs != 0)
        {
            push(@$tbox, "\\mathop{\\geq $_} $rn \\sqsubseteq \\mathop{\\geq $qs} $rn");
            #print "adding axiom: \\mathop{\\geq $_} $rn \\sqsubseteq \\mathop{\\geq $qs} $rn\n";
        }
        $qs = $_;
    }

    @q - 1;
}
